#include <postgres.h>
#include <fmgr.h>

#include "uint.h"

#include <utils/sortsupport.h>


PG_FUNCTION_INFO_V1(int1int1lt);
Datum
int1int1lt(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int1int1le);
Datum
int1int1le(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int1int1eq);
Datum
int1int1eq(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int1int1ne);
Datum
int1int1ne(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int1int1ge);
Datum
int1int1ge(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int1int1gt);
Datum
int1int1gt(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btint1int1cmp);
Datum
btint1int1cmp(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int8 arg2 = PG_GETARG_INT8(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int1int1pl);
Datum
int1int1pl(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int8 arg2 = PG_GETARG_INT8(1);
	int8 result;

	result = arg1 + arg2;

	if (SAMESIGN(arg1, arg2) && !SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT8(result);
}

PG_FUNCTION_INFO_V1(int1int1mi);
Datum
int1int1mi(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int8 arg2 = PG_GETARG_INT8(1);
	int8 result;

	result = arg1 - arg2;

	if (!SAMESIGN(arg1, arg2) && !SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT8(result);
}

PG_FUNCTION_INFO_V1(int1int1mul);
Datum
int1int1mul(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int8 arg2 = PG_GETARG_INT8(1);
	int8 result;

	int16 result2;

	result2 = (int16)arg1 * (int16)arg2;

	if ((int8) result2 != result2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));
	result = result2;

	PG_RETURN_INT8(result);
}

PG_FUNCTION_INFO_V1(int1int1div);
Datum
int1int1div(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int8 arg2 = PG_GETARG_INT8(1);
	int8 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	PG_RETURN_INT8(result);
}

PG_FUNCTION_INFO_V1(int1int1mod);
Datum
int1int1mod(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int8 arg2 = PG_GETARG_INT8(1);
	int8 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	if (arg2 == -1)
		PG_RETURN_INT8(0);

	result = arg1 % arg2;

	PG_RETURN_INT8(result);
}

PG_FUNCTION_INFO_V1(int1uint1lt);
Datum
int1uint1lt(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int1uint1le);
Datum
int1uint1le(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int1uint1eq);
Datum
int1uint1eq(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int1uint1ne);
Datum
int1uint1ne(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int1uint1ge);
Datum
int1uint1ge(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int1uint1gt);
Datum
int1uint1gt(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btint1uint1cmp);
Datum
btint1uint1cmp(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int1uint1pl);
Datum
int1uint1pl(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	uint8 result;

	result = arg1 + arg2;

	if ((arg1 < 0 && result > arg2) || (arg1 > 0 && result < arg2))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT8(result);
}

PG_FUNCTION_INFO_V1(int1uint1mi);
Datum
int1uint1mi(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	uint8 result;

	result = arg1 - arg2;

	if ((arg1 < 0) || (result > arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT8(result);
}

PG_FUNCTION_INFO_V1(int1uint1mul);
Datum
int1uint1mul(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	uint8 result;

	uint16 result2;

	result2 = (uint16)arg1 * (uint16)arg2;

	if ((uint8) result2 != result2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));
	result = result2;

	PG_RETURN_UINT8(result);
}

PG_FUNCTION_INFO_V1(int1uint1div);
Datum
int1uint1div(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	uint8 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	if (arg1 < 0)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT8(result);
}

PG_FUNCTION_INFO_V1(int1uint1mod);
Datum
int1uint1mod(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	uint8 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 % arg2;

	if (arg1 < 0)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT8(result);
}

PG_FUNCTION_INFO_V1(int1_to_uint1);
Datum
int1_to_uint1(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint8 result;

	result = arg1;
	if ((int8) result != arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("uint1 out of range")));
	if (!SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("uint1 out of range")));

	PG_RETURN_UINT8(result);
}

PG_FUNCTION_INFO_V1(int1uint2lt);
Datum
int1uint2lt(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int1uint2le);
Datum
int1uint2le(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int1uint2eq);
Datum
int1uint2eq(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int1uint2ne);
Datum
int1uint2ne(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int1uint2ge);
Datum
int1uint2ge(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int1uint2gt);
Datum
int1uint2gt(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btint1uint2cmp);
Datum
btint1uint2cmp(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int1uint2pl);
Datum
int1uint2pl(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	uint16 result;

	result = arg1 + arg2;

	if ((arg1 < 0 && result > arg2) || (arg1 > 0 && result < arg2))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(int1uint2mi);
Datum
int1uint2mi(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	uint16 result;

	result = arg1 - arg2;

	if ((arg1 < 0) || (result > arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(int1uint2mul);
Datum
int1uint2mul(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	uint16 result;

	uint32 result2;

	result2 = (uint32)arg1 * (uint32)arg2;

	if ((uint16) result2 != result2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));
	result = result2;

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(int1uint2div);
Datum
int1uint2div(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	uint16 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	if (arg1 < 0)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(int1uint2mod);
Datum
int1uint2mod(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	uint16 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 % arg2;

	if (arg1 < 0)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(int1_to_uint2);
Datum
int1_to_uint2(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint16 result;

	result = arg1;
	if (!SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("uint2 out of range")));

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(int1uint4lt);
Datum
int1uint4lt(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int1uint4le);
Datum
int1uint4le(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int1uint4eq);
Datum
int1uint4eq(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int1uint4ne);
Datum
int1uint4ne(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int1uint4ge);
Datum
int1uint4ge(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int1uint4gt);
Datum
int1uint4gt(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btint1uint4cmp);
Datum
btint1uint4cmp(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int1uint4pl);
Datum
int1uint4pl(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	uint32 result;

	result = arg1 + arg2;

	if ((arg1 < 0 && result > arg2) || (arg1 > 0 && result < arg2))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(int1uint4mi);
Datum
int1uint4mi(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	uint32 result;

	result = arg1 - arg2;

	if ((arg1 < 0) || (result > arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(int1uint4mul);
Datum
int1uint4mul(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	uint32 result;

	uint64 result2;

	result2 = (uint64)arg1 * (uint64)arg2;

	if ((uint32) result2 != result2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));
	result = result2;

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(int1uint4div);
Datum
int1uint4div(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	uint32 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	if (arg1 < 0)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(int1uint4mod);
Datum
int1uint4mod(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	uint32 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 % arg2;

	if (arg1 < 0)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(int1_to_uint4);
Datum
int1_to_uint4(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint32 result;

	result = arg1;
	if (!SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("uint4 out of range")));

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(int1uint8lt);
Datum
int1uint8lt(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int1uint8le);
Datum
int1uint8le(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int1uint8eq);
Datum
int1uint8eq(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int1uint8ne);
Datum
int1uint8ne(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int1uint8ge);
Datum
int1uint8ge(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int1uint8gt);
Datum
int1uint8gt(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btint1uint8cmp);
Datum
btint1uint8cmp(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int1uint8pl);
Datum
int1uint8pl(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	result = arg1 + arg2;

	if ((arg1 < 0 && result > arg2) || (arg1 > 0 && result < arg2))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(int1uint8mi);
Datum
int1uint8mi(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	result = arg1 - arg2;

	if ((arg1 < 0) || (result > arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(int1uint8mul);
Datum
int1uint8mul(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	result = arg1 * arg2;

	if ((arg1 < 0) || ((arg1 != ((uint32) arg1) || arg2 != ((uint32) arg2)) && (arg2 != 0 && (result / arg2 != arg1))))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(int1uint8div);
Datum
int1uint8div(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	if (arg1 < 0)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(int1uint8mod);
Datum
int1uint8mod(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 % arg2;

	if (arg1 < 0)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(int1_to_uint8);
Datum
int1_to_uint8(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	uint64 result;

	result = arg1;
	if (!SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("uint8 out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(int1int2lt);
Datum
int1int2lt(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int16 arg2 = PG_GETARG_INT16(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int1int2le);
Datum
int1int2le(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int16 arg2 = PG_GETARG_INT16(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int1int2eq);
Datum
int1int2eq(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int16 arg2 = PG_GETARG_INT16(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int1int2ne);
Datum
int1int2ne(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int16 arg2 = PG_GETARG_INT16(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int1int2ge);
Datum
int1int2ge(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int16 arg2 = PG_GETARG_INT16(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int1int2gt);
Datum
int1int2gt(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int16 arg2 = PG_GETARG_INT16(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btint1int2cmp);
Datum
btint1int2cmp(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int16 arg2 = PG_GETARG_INT16(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int1int2pl);
Datum
int1int2pl(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int16 arg2 = PG_GETARG_INT16(1);
	int16 result;

	result = arg1 + arg2;

	if (SAMESIGN(arg1, arg2) && !SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT16(result);
}

PG_FUNCTION_INFO_V1(int1int2mi);
Datum
int1int2mi(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int16 arg2 = PG_GETARG_INT16(1);
	int16 result;

	result = arg1 - arg2;

	if (!SAMESIGN(arg1, arg2) && !SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT16(result);
}

PG_FUNCTION_INFO_V1(int1int2mul);
Datum
int1int2mul(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int16 arg2 = PG_GETARG_INT16(1);
	int16 result;

	int32 result2;

	result2 = (int32)arg1 * (int32)arg2;

	if ((int16) result2 != result2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));
	result = result2;

	PG_RETURN_INT16(result);
}

PG_FUNCTION_INFO_V1(int1int2div);
Datum
int1int2div(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int16 arg2 = PG_GETARG_INT16(1);
	int16 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	PG_RETURN_INT16(result);
}

PG_FUNCTION_INFO_V1(int1int2mod);
Datum
int1int2mod(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int16 arg2 = PG_GETARG_INT16(1);
	int16 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	if (arg2 == -1)
		PG_RETURN_INT16(0);

	result = arg1 % arg2;

	PG_RETURN_INT16(result);
}

PG_FUNCTION_INFO_V1(int1_to_int2);
Datum
int1_to_int2(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int16 result;

	result = arg1;

	PG_RETURN_INT16(result);
}

PG_FUNCTION_INFO_V1(int1int4lt);
Datum
int1int4lt(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int32 arg2 = PG_GETARG_INT32(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int1int4le);
Datum
int1int4le(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int32 arg2 = PG_GETARG_INT32(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int1int4eq);
Datum
int1int4eq(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int32 arg2 = PG_GETARG_INT32(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int1int4ne);
Datum
int1int4ne(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int32 arg2 = PG_GETARG_INT32(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int1int4ge);
Datum
int1int4ge(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int32 arg2 = PG_GETARG_INT32(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int1int4gt);
Datum
int1int4gt(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int32 arg2 = PG_GETARG_INT32(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btint1int4cmp);
Datum
btint1int4cmp(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int32 arg2 = PG_GETARG_INT32(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int1int4pl);
Datum
int1int4pl(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int32 arg2 = PG_GETARG_INT32(1);
	int32 result;

	result = arg1 + arg2;

	if (SAMESIGN(arg1, arg2) && !SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int1int4mi);
Datum
int1int4mi(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int32 arg2 = PG_GETARG_INT32(1);
	int32 result;

	result = arg1 - arg2;

	if (!SAMESIGN(arg1, arg2) && !SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int1int4mul);
Datum
int1int4mul(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int32 arg2 = PG_GETARG_INT32(1);
	int32 result;

	int64 result2;

	result2 = (int64)arg1 * (int64)arg2;

	if ((int32) result2 != result2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));
	result = result2;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int1int4div);
Datum
int1int4div(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int32 arg2 = PG_GETARG_INT32(1);
	int32 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int1int4mod);
Datum
int1int4mod(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int32 arg2 = PG_GETARG_INT32(1);
	int32 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	if (arg2 == -1)
		PG_RETURN_INT32(0);

	result = arg1 % arg2;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int1_to_int4);
Datum
int1_to_int4(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int32 result;

	result = arg1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int1int8lt);
Datum
int1int8lt(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int64 arg2 = PG_GETARG_INT64(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int1int8le);
Datum
int1int8le(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int64 arg2 = PG_GETARG_INT64(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int1int8eq);
Datum
int1int8eq(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int64 arg2 = PG_GETARG_INT64(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int1int8ne);
Datum
int1int8ne(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int64 arg2 = PG_GETARG_INT64(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int1int8ge);
Datum
int1int8ge(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int64 arg2 = PG_GETARG_INT64(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int1int8gt);
Datum
int1int8gt(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int64 arg2 = PG_GETARG_INT64(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btint1int8cmp);
Datum
btint1int8cmp(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int64 arg2 = PG_GETARG_INT64(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int1int8pl);
Datum
int1int8pl(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int64 arg2 = PG_GETARG_INT64(1);
	int64 result;

	result = arg1 + arg2;

	if (SAMESIGN(arg1, arg2) && !SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(int1int8mi);
Datum
int1int8mi(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int64 arg2 = PG_GETARG_INT64(1);
	int64 result;

	result = arg1 - arg2;

	if (!SAMESIGN(arg1, arg2) && !SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(int1int8mul);
Datum
int1int8mul(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int64 arg2 = PG_GETARG_INT64(1);
	int64 result;

	result = arg1 * arg2;

	if ((arg1 != ((int32) arg1) || arg2 != ((int32) arg2)) && (arg2 != 0 && ((arg2 == -1 && arg1 < 0 && result < 0) || result / arg2 != arg1)))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(int1int8div);
Datum
int1int8div(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int64 arg2 = PG_GETARG_INT64(1);
	int64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(int1int8mod);
Datum
int1int8mod(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int64 arg2 = PG_GETARG_INT64(1);
	int64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	if (arg2 == -1)
		PG_RETURN_INT64(0);

	result = arg1 % arg2;

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(int1_to_int8);
Datum
int1_to_int8(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int64 result;

	result = arg1;

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(uint1int1lt);
Datum
uint1int1lt(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint1int1le);
Datum
uint1int1le(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint1int1eq);
Datum
uint1int1eq(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint1int1ne);
Datum
uint1int1ne(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint1int1ge);
Datum
uint1int1ge(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint1int1gt);
Datum
uint1int1gt(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btuint1int1cmp);
Datum
btuint1int1cmp(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int8 arg2 = PG_GETARG_INT8(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint1int1pl);
Datum
uint1int1pl(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int8 arg2 = PG_GETARG_INT8(1);
	uint8 result;

	result = arg1 + arg2;

	if ((arg2 < 0 && result > arg1) || (arg2 > 0 && result < arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT8(result);
}

PG_FUNCTION_INFO_V1(uint1int1mi);
Datum
uint1int1mi(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int8 arg2 = PG_GETARG_INT8(1);
	uint8 result;

	result = arg1 - arg2;

	if ((arg2 < 0 && result < arg1) || (arg2 > 0 && result > arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT8(result);
}

PG_FUNCTION_INFO_V1(uint1int1mul);
Datum
uint1int1mul(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int8 arg2 = PG_GETARG_INT8(1);
	uint8 result;

	uint16 result2;

	result2 = (uint16)arg1 * (uint16)arg2;

	if ((uint8) result2 != result2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));
	result = result2;

	PG_RETURN_UINT8(result);
}

PG_FUNCTION_INFO_V1(uint1int1div);
Datum
uint1int1div(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int8 arg2 = PG_GETARG_INT8(1);
	uint8 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	if (arg2 < 0)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT8(result);
}

PG_FUNCTION_INFO_V1(uint1int1mod);
Datum
uint1int1mod(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int8 arg2 = PG_GETARG_INT8(1);
	uint8 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	if (arg2 == -1)
		PG_RETURN_UINT8(0);

	result = arg1 % arg2;

	if (arg2 < 0)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT8(result);
}

PG_FUNCTION_INFO_V1(uint1_to_int1);
Datum
uint1_to_int1(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int8 result;

	result = arg1;
	if ((uint8) result != arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("int1 out of range")));
	if (!SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("int1 out of range")));

	PG_RETURN_INT8(result);
}

PG_FUNCTION_INFO_V1(uint1uint1lt);
Datum
uint1uint1lt(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint1uint1le);
Datum
uint1uint1le(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint1uint1eq);
Datum
uint1uint1eq(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint1uint1ne);
Datum
uint1uint1ne(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint1uint1ge);
Datum
uint1uint1ge(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint1uint1gt);
Datum
uint1uint1gt(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btuint1uint1cmp);
Datum
btuint1uint1cmp(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint1uint1pl);
Datum
uint1uint1pl(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	uint8 result;

	result = arg1 + arg2;

	if (result < arg1 || result < arg2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT8(result);
}

PG_FUNCTION_INFO_V1(uint1uint1mi);
Datum
uint1uint1mi(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	uint8 result;

	result = arg1 - arg2;

	if (result > arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT8(result);
}

PG_FUNCTION_INFO_V1(uint1uint1mul);
Datum
uint1uint1mul(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	uint8 result;

	uint16 result2;

	result2 = (uint16)arg1 * (uint16)arg2;

	if ((uint8) result2 != result2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));
	result = result2;

	PG_RETURN_UINT8(result);
}

PG_FUNCTION_INFO_V1(uint1uint1div);
Datum
uint1uint1div(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	uint8 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	PG_RETURN_UINT8(result);
}

PG_FUNCTION_INFO_V1(uint1uint1mod);
Datum
uint1uint1mod(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	uint8 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 % arg2;

	PG_RETURN_UINT8(result);
}

PG_FUNCTION_INFO_V1(uint1uint2lt);
Datum
uint1uint2lt(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint1uint2le);
Datum
uint1uint2le(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint1uint2eq);
Datum
uint1uint2eq(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint1uint2ne);
Datum
uint1uint2ne(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint1uint2ge);
Datum
uint1uint2ge(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint1uint2gt);
Datum
uint1uint2gt(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btuint1uint2cmp);
Datum
btuint1uint2cmp(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint1uint2pl);
Datum
uint1uint2pl(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	uint16 result;

	result = arg1 + arg2;

	if (result < arg1 || result < arg2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(uint1uint2mi);
Datum
uint1uint2mi(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	uint16 result;

	result = arg1 - arg2;

	if (result > arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(uint1uint2mul);
Datum
uint1uint2mul(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	uint16 result;

	uint32 result2;

	result2 = (uint32)arg1 * (uint32)arg2;

	if ((uint16) result2 != result2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));
	result = result2;

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(uint1uint2div);
Datum
uint1uint2div(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	uint16 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(uint1uint2mod);
Datum
uint1uint2mod(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	uint16 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 % arg2;

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(uint1_to_uint2);
Datum
uint1_to_uint2(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint16 result;

	result = arg1;

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(uint1uint4lt);
Datum
uint1uint4lt(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint1uint4le);
Datum
uint1uint4le(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint1uint4eq);
Datum
uint1uint4eq(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint1uint4ne);
Datum
uint1uint4ne(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint1uint4ge);
Datum
uint1uint4ge(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint1uint4gt);
Datum
uint1uint4gt(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btuint1uint4cmp);
Datum
btuint1uint4cmp(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint1uint4pl);
Datum
uint1uint4pl(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	uint32 result;

	result = arg1 + arg2;

	if (result < arg1 || result < arg2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint1uint4mi);
Datum
uint1uint4mi(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	uint32 result;

	result = arg1 - arg2;

	if (result > arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint1uint4mul);
Datum
uint1uint4mul(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	uint32 result;

	uint64 result2;

	result2 = (uint64)arg1 * (uint64)arg2;

	if ((uint32) result2 != result2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));
	result = result2;

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint1uint4div);
Datum
uint1uint4div(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	uint32 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint1uint4mod);
Datum
uint1uint4mod(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	uint32 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 % arg2;

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint1_to_uint4);
Datum
uint1_to_uint4(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint32 result;

	result = arg1;

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint1uint8lt);
Datum
uint1uint8lt(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint1uint8le);
Datum
uint1uint8le(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint1uint8eq);
Datum
uint1uint8eq(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint1uint8ne);
Datum
uint1uint8ne(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint1uint8ge);
Datum
uint1uint8ge(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint1uint8gt);
Datum
uint1uint8gt(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btuint1uint8cmp);
Datum
btuint1uint8cmp(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint1uint8pl);
Datum
uint1uint8pl(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	result = arg1 + arg2;

	if (result < arg1 || result < arg2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint1uint8mi);
Datum
uint1uint8mi(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	result = arg1 - arg2;

	if (result > arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint1uint8mul);
Datum
uint1uint8mul(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	result = arg1 * arg2;

	if (((arg1 != ((uint32) arg1) || arg2 != ((uint32) arg2)) && (arg2 != 0 && (result / arg2 != arg1))))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint1uint8div);
Datum
uint1uint8div(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint1uint8mod);
Datum
uint1uint8mod(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 % arg2;

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint1_to_uint8);
Datum
uint1_to_uint8(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint64 result;

	result = arg1;

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint1int2lt);
Datum
uint1int2lt(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int16 arg2 = PG_GETARG_INT16(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint1int2le);
Datum
uint1int2le(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int16 arg2 = PG_GETARG_INT16(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint1int2eq);
Datum
uint1int2eq(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int16 arg2 = PG_GETARG_INT16(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint1int2ne);
Datum
uint1int2ne(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int16 arg2 = PG_GETARG_INT16(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint1int2ge);
Datum
uint1int2ge(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int16 arg2 = PG_GETARG_INT16(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint1int2gt);
Datum
uint1int2gt(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int16 arg2 = PG_GETARG_INT16(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btuint1int2cmp);
Datum
btuint1int2cmp(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int16 arg2 = PG_GETARG_INT16(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint1int2pl);
Datum
uint1int2pl(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int16 arg2 = PG_GETARG_INT16(1);
	int16 result;

	result = arg1 + arg2;

	if (SAMESIGN(arg1, arg2) && !SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT16(result);
}

PG_FUNCTION_INFO_V1(uint1int2mi);
Datum
uint1int2mi(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int16 arg2 = PG_GETARG_INT16(1);
	int16 result;

	result = arg1 - arg2;

	if (!SAMESIGN(arg1, arg2) && !SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT16(result);
}

PG_FUNCTION_INFO_V1(uint1int2mul);
Datum
uint1int2mul(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int16 arg2 = PG_GETARG_INT16(1);
	int16 result;

	int32 result2;

	result2 = (int32)arg1 * (int32)arg2;

	if ((int16) result2 != result2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));
	result = result2;

	PG_RETURN_INT16(result);
}

PG_FUNCTION_INFO_V1(uint1int2div);
Datum
uint1int2div(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int16 arg2 = PG_GETARG_INT16(1);
	int16 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	PG_RETURN_INT16(result);
}

PG_FUNCTION_INFO_V1(uint1int2mod);
Datum
uint1int2mod(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int16 arg2 = PG_GETARG_INT16(1);
	int16 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	if (arg2 == -1)
		PG_RETURN_INT16(0);

	result = arg1 % arg2;

	PG_RETURN_INT16(result);
}

PG_FUNCTION_INFO_V1(uint1_to_int2);
Datum
uint1_to_int2(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int16 result;

	result = arg1;
	if (!SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("int2 out of range")));

	PG_RETURN_INT16(result);
}

PG_FUNCTION_INFO_V1(uint1int4lt);
Datum
uint1int4lt(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int32 arg2 = PG_GETARG_INT32(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint1int4le);
Datum
uint1int4le(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int32 arg2 = PG_GETARG_INT32(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint1int4eq);
Datum
uint1int4eq(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int32 arg2 = PG_GETARG_INT32(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint1int4ne);
Datum
uint1int4ne(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int32 arg2 = PG_GETARG_INT32(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint1int4ge);
Datum
uint1int4ge(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int32 arg2 = PG_GETARG_INT32(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint1int4gt);
Datum
uint1int4gt(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int32 arg2 = PG_GETARG_INT32(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btuint1int4cmp);
Datum
btuint1int4cmp(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int32 arg2 = PG_GETARG_INT32(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint1int4pl);
Datum
uint1int4pl(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int32 arg2 = PG_GETARG_INT32(1);
	int32 result;

	result = arg1 + arg2;

	if (SAMESIGN(arg1, arg2) && !SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint1int4mi);
Datum
uint1int4mi(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int32 arg2 = PG_GETARG_INT32(1);
	int32 result;

	result = arg1 - arg2;

	if (!SAMESIGN(arg1, arg2) && !SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint1int4mul);
Datum
uint1int4mul(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int32 arg2 = PG_GETARG_INT32(1);
	int32 result;

	int64 result2;

	result2 = (int64)arg1 * (int64)arg2;

	if ((int32) result2 != result2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));
	result = result2;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint1int4div);
Datum
uint1int4div(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int32 arg2 = PG_GETARG_INT32(1);
	int32 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint1int4mod);
Datum
uint1int4mod(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int32 arg2 = PG_GETARG_INT32(1);
	int32 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	if (arg2 == -1)
		PG_RETURN_INT32(0);

	result = arg1 % arg2;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint1_to_int4);
Datum
uint1_to_int4(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int32 result;

	result = arg1;
	if (!SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("int4 out of range")));

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint1int8lt);
Datum
uint1int8lt(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int64 arg2 = PG_GETARG_INT64(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint1int8le);
Datum
uint1int8le(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int64 arg2 = PG_GETARG_INT64(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint1int8eq);
Datum
uint1int8eq(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int64 arg2 = PG_GETARG_INT64(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint1int8ne);
Datum
uint1int8ne(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int64 arg2 = PG_GETARG_INT64(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint1int8ge);
Datum
uint1int8ge(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int64 arg2 = PG_GETARG_INT64(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint1int8gt);
Datum
uint1int8gt(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int64 arg2 = PG_GETARG_INT64(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btuint1int8cmp);
Datum
btuint1int8cmp(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int64 arg2 = PG_GETARG_INT64(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint1int8pl);
Datum
uint1int8pl(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int64 arg2 = PG_GETARG_INT64(1);
	int64 result;

	result = arg1 + arg2;

	if (SAMESIGN(arg1, arg2) && !SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(uint1int8mi);
Datum
uint1int8mi(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int64 arg2 = PG_GETARG_INT64(1);
	int64 result;

	result = arg1 - arg2;

	if (!SAMESIGN(arg1, arg2) && !SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(uint1int8mul);
Datum
uint1int8mul(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int64 arg2 = PG_GETARG_INT64(1);
	int64 result;

	result = arg1 * arg2;

	if ((arg1 != ((int32) arg1) || arg2 != ((int32) arg2)) && (arg2 != 0 && (result / arg2 != arg1)))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(uint1int8div);
Datum
uint1int8div(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int64 arg2 = PG_GETARG_INT64(1);
	int64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(uint1int8mod);
Datum
uint1int8mod(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int64 arg2 = PG_GETARG_INT64(1);
	int64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	if (arg2 == -1)
		PG_RETURN_INT64(0);

	result = arg1 % arg2;

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(uint1_to_int8);
Datum
uint1_to_int8(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int64 result;

	result = arg1;
	if (!SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("int8 out of range")));

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(uint2int1lt);
Datum
uint2int1lt(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint2int1le);
Datum
uint2int1le(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint2int1eq);
Datum
uint2int1eq(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint2int1ne);
Datum
uint2int1ne(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint2int1ge);
Datum
uint2int1ge(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint2int1gt);
Datum
uint2int1gt(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btuint2int1cmp);
Datum
btuint2int1cmp(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int8 arg2 = PG_GETARG_INT8(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint2int1pl);
Datum
uint2int1pl(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int8 arg2 = PG_GETARG_INT8(1);
	uint16 result;

	result = arg1 + arg2;

	if ((arg2 < 0 && result > arg1) || (arg2 > 0 && result < arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(uint2int1mi);
Datum
uint2int1mi(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int8 arg2 = PG_GETARG_INT8(1);
	uint16 result;

	result = arg1 - arg2;

	if ((arg2 < 0 && result < arg1) || (arg2 > 0 && result > arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(uint2int1mul);
Datum
uint2int1mul(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int8 arg2 = PG_GETARG_INT8(1);
	uint16 result;

	uint32 result2;

	result2 = (uint32)arg1 * (uint32)arg2;

	if ((uint16) result2 != result2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));
	result = result2;

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(uint2int1div);
Datum
uint2int1div(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int8 arg2 = PG_GETARG_INT8(1);
	uint16 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	if (arg2 < 0)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(uint2int1mod);
Datum
uint2int1mod(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int8 arg2 = PG_GETARG_INT8(1);
	uint16 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	if (arg2 == -1)
		PG_RETURN_UINT16(0);

	result = arg1 % arg2;

	if (arg2 < 0)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(uint2_to_int1);
Datum
uint2_to_int1(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int8 result;

	result = arg1;
	if ((uint16) result != arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("int1 out of range")));
	if (!SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("int1 out of range")));

	PG_RETURN_INT8(result);
}

PG_FUNCTION_INFO_V1(uint2uint1lt);
Datum
uint2uint1lt(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint2uint1le);
Datum
uint2uint1le(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint2uint1eq);
Datum
uint2uint1eq(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint2uint1ne);
Datum
uint2uint1ne(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint2uint1ge);
Datum
uint2uint1ge(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint2uint1gt);
Datum
uint2uint1gt(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btuint2uint1cmp);
Datum
btuint2uint1cmp(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint2uint1pl);
Datum
uint2uint1pl(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	uint16 result;

	result = arg1 + arg2;

	if (result < arg1 || result < arg2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(uint2uint1mi);
Datum
uint2uint1mi(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	uint16 result;

	result = arg1 - arg2;

	if (result > arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(uint2uint1mul);
Datum
uint2uint1mul(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	uint16 result;

	uint32 result2;

	result2 = (uint32)arg1 * (uint32)arg2;

	if ((uint16) result2 != result2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));
	result = result2;

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(uint2uint1div);
Datum
uint2uint1div(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	uint16 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(uint2uint1mod);
Datum
uint2uint1mod(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	uint16 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 % arg2;

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(uint2_to_uint1);
Datum
uint2_to_uint1(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint8 result;

	result = arg1;
	if ((uint16) result != arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("uint1 out of range")));

	PG_RETURN_UINT8(result);
}

PG_FUNCTION_INFO_V1(uint2uint2lt);
Datum
uint2uint2lt(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint2uint2le);
Datum
uint2uint2le(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint2uint2eq);
Datum
uint2uint2eq(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint2uint2ne);
Datum
uint2uint2ne(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint2uint2ge);
Datum
uint2uint2ge(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint2uint2gt);
Datum
uint2uint2gt(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btuint2uint2cmp);
Datum
btuint2uint2cmp(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint2uint2pl);
Datum
uint2uint2pl(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	uint16 result;

	result = arg1 + arg2;

	if (result < arg1 || result < arg2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(uint2uint2mi);
Datum
uint2uint2mi(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	uint16 result;

	result = arg1 - arg2;

	if (result > arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(uint2uint2mul);
Datum
uint2uint2mul(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	uint16 result;

	uint32 result2;

	result2 = (uint32)arg1 * (uint32)arg2;

	if ((uint16) result2 != result2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));
	result = result2;

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(uint2uint2div);
Datum
uint2uint2div(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	uint16 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(uint2uint2mod);
Datum
uint2uint2mod(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	uint16 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 % arg2;

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(uint2uint4lt);
Datum
uint2uint4lt(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint2uint4le);
Datum
uint2uint4le(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint2uint4eq);
Datum
uint2uint4eq(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint2uint4ne);
Datum
uint2uint4ne(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint2uint4ge);
Datum
uint2uint4ge(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint2uint4gt);
Datum
uint2uint4gt(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btuint2uint4cmp);
Datum
btuint2uint4cmp(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint2uint4pl);
Datum
uint2uint4pl(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	uint32 result;

	result = arg1 + arg2;

	if (result < arg1 || result < arg2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint2uint4mi);
Datum
uint2uint4mi(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	uint32 result;

	result = arg1 - arg2;

	if (result > arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint2uint4mul);
Datum
uint2uint4mul(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	uint32 result;

	uint64 result2;

	result2 = (uint64)arg1 * (uint64)arg2;

	if ((uint32) result2 != result2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));
	result = result2;

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint2uint4div);
Datum
uint2uint4div(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	uint32 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint2uint4mod);
Datum
uint2uint4mod(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	uint32 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 % arg2;

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint2_to_uint4);
Datum
uint2_to_uint4(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint32 result;

	result = arg1;

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint2uint8lt);
Datum
uint2uint8lt(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint2uint8le);
Datum
uint2uint8le(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint2uint8eq);
Datum
uint2uint8eq(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint2uint8ne);
Datum
uint2uint8ne(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint2uint8ge);
Datum
uint2uint8ge(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint2uint8gt);
Datum
uint2uint8gt(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btuint2uint8cmp);
Datum
btuint2uint8cmp(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint2uint8pl);
Datum
uint2uint8pl(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	result = arg1 + arg2;

	if (result < arg1 || result < arg2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint2uint8mi);
Datum
uint2uint8mi(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	result = arg1 - arg2;

	if (result > arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint2uint8mul);
Datum
uint2uint8mul(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	result = arg1 * arg2;

	if (((arg1 != ((uint32) arg1) || arg2 != ((uint32) arg2)) && (arg2 != 0 && (result / arg2 != arg1))))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint2uint8div);
Datum
uint2uint8div(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint2uint8mod);
Datum
uint2uint8mod(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 % arg2;

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint2_to_uint8);
Datum
uint2_to_uint8(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint64 result;

	result = arg1;

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint2int2lt);
Datum
uint2int2lt(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int16 arg2 = PG_GETARG_INT16(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint2int2le);
Datum
uint2int2le(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int16 arg2 = PG_GETARG_INT16(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint2int2eq);
Datum
uint2int2eq(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int16 arg2 = PG_GETARG_INT16(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint2int2ne);
Datum
uint2int2ne(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int16 arg2 = PG_GETARG_INT16(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint2int2ge);
Datum
uint2int2ge(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int16 arg2 = PG_GETARG_INT16(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint2int2gt);
Datum
uint2int2gt(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int16 arg2 = PG_GETARG_INT16(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btuint2int2cmp);
Datum
btuint2int2cmp(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int16 arg2 = PG_GETARG_INT16(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint2int2pl);
Datum
uint2int2pl(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int16 arg2 = PG_GETARG_INT16(1);
	uint16 result;

	result = arg1 + arg2;

	if ((arg2 < 0 && result > arg1) || (arg2 > 0 && result < arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(uint2int2mi);
Datum
uint2int2mi(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int16 arg2 = PG_GETARG_INT16(1);
	uint16 result;

	result = arg1 - arg2;

	if ((arg2 < 0 && result < arg1) || (arg2 > 0 && result > arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(uint2int2mul);
Datum
uint2int2mul(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int16 arg2 = PG_GETARG_INT16(1);
	uint16 result;

	uint32 result2;

	result2 = (uint32)arg1 * (uint32)arg2;

	if ((uint16) result2 != result2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));
	result = result2;

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(uint2int2div);
Datum
uint2int2div(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int16 arg2 = PG_GETARG_INT16(1);
	uint16 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	if (arg2 < 0)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(uint2int2mod);
Datum
uint2int2mod(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int16 arg2 = PG_GETARG_INT16(1);
	uint16 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	if (arg2 == -1)
		PG_RETURN_UINT16(0);

	result = arg1 % arg2;

	if (arg2 < 0)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(uint2_to_int2);
Datum
uint2_to_int2(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int16 result;

	result = arg1;
	if ((uint16) result != arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("int2 out of range")));
	if (!SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("int2 out of range")));

	PG_RETURN_INT16(result);
}

PG_FUNCTION_INFO_V1(uint2int4lt);
Datum
uint2int4lt(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int32 arg2 = PG_GETARG_INT32(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint2int4le);
Datum
uint2int4le(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int32 arg2 = PG_GETARG_INT32(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint2int4eq);
Datum
uint2int4eq(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int32 arg2 = PG_GETARG_INT32(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint2int4ne);
Datum
uint2int4ne(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int32 arg2 = PG_GETARG_INT32(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint2int4ge);
Datum
uint2int4ge(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int32 arg2 = PG_GETARG_INT32(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint2int4gt);
Datum
uint2int4gt(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int32 arg2 = PG_GETARG_INT32(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btuint2int4cmp);
Datum
btuint2int4cmp(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int32 arg2 = PG_GETARG_INT32(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint2int4pl);
Datum
uint2int4pl(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int32 arg2 = PG_GETARG_INT32(1);
	int32 result;

	result = arg1 + arg2;

	if (SAMESIGN(arg1, arg2) && !SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint2int4mi);
Datum
uint2int4mi(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int32 arg2 = PG_GETARG_INT32(1);
	int32 result;

	result = arg1 - arg2;

	if (!SAMESIGN(arg1, arg2) && !SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint2int4mul);
Datum
uint2int4mul(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int32 arg2 = PG_GETARG_INT32(1);
	int32 result;

	int64 result2;

	result2 = (int64)arg1 * (int64)arg2;

	if ((int32) result2 != result2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));
	result = result2;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint2int4div);
Datum
uint2int4div(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int32 arg2 = PG_GETARG_INT32(1);
	int32 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint2int4mod);
Datum
uint2int4mod(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int32 arg2 = PG_GETARG_INT32(1);
	int32 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	if (arg2 == -1)
		PG_RETURN_INT32(0);

	result = arg1 % arg2;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint2_to_int4);
Datum
uint2_to_int4(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int32 result;

	result = arg1;
	if (!SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("int4 out of range")));

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint2int8lt);
Datum
uint2int8lt(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int64 arg2 = PG_GETARG_INT64(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint2int8le);
Datum
uint2int8le(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int64 arg2 = PG_GETARG_INT64(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint2int8eq);
Datum
uint2int8eq(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int64 arg2 = PG_GETARG_INT64(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint2int8ne);
Datum
uint2int8ne(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int64 arg2 = PG_GETARG_INT64(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint2int8ge);
Datum
uint2int8ge(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int64 arg2 = PG_GETARG_INT64(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint2int8gt);
Datum
uint2int8gt(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int64 arg2 = PG_GETARG_INT64(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btuint2int8cmp);
Datum
btuint2int8cmp(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int64 arg2 = PG_GETARG_INT64(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint2int8pl);
Datum
uint2int8pl(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int64 arg2 = PG_GETARG_INT64(1);
	int64 result;

	result = arg1 + arg2;

	if (SAMESIGN(arg1, arg2) && !SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(uint2int8mi);
Datum
uint2int8mi(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int64 arg2 = PG_GETARG_INT64(1);
	int64 result;

	result = arg1 - arg2;

	if (!SAMESIGN(arg1, arg2) && !SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(uint2int8mul);
Datum
uint2int8mul(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int64 arg2 = PG_GETARG_INT64(1);
	int64 result;

	result = arg1 * arg2;

	if ((arg1 != ((int32) arg1) || arg2 != ((int32) arg2)) && (arg2 != 0 && (result / arg2 != arg1)))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(uint2int8div);
Datum
uint2int8div(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int64 arg2 = PG_GETARG_INT64(1);
	int64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(uint2int8mod);
Datum
uint2int8mod(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int64 arg2 = PG_GETARG_INT64(1);
	int64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	if (arg2 == -1)
		PG_RETURN_INT64(0);

	result = arg1 % arg2;

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(uint2_to_int8);
Datum
uint2_to_int8(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int64 result;

	result = arg1;
	if (!SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("int8 out of range")));

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(uint4int1lt);
Datum
uint4int1lt(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint4int1le);
Datum
uint4int1le(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint4int1eq);
Datum
uint4int1eq(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint4int1ne);
Datum
uint4int1ne(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint4int1ge);
Datum
uint4int1ge(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint4int1gt);
Datum
uint4int1gt(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btuint4int1cmp);
Datum
btuint4int1cmp(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int8 arg2 = PG_GETARG_INT8(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint4int1pl);
Datum
uint4int1pl(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int8 arg2 = PG_GETARG_INT8(1);
	uint32 result;

	result = arg1 + arg2;

	if ((arg2 < 0 && result > arg1) || (arg2 > 0 && result < arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint4int1mi);
Datum
uint4int1mi(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int8 arg2 = PG_GETARG_INT8(1);
	uint32 result;

	result = arg1 - arg2;

	if ((arg2 < 0 && result < arg1) || (arg2 > 0 && result > arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint4int1mul);
Datum
uint4int1mul(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int8 arg2 = PG_GETARG_INT8(1);
	uint32 result;

	uint64 result2;

	result2 = (uint64)arg1 * (uint64)arg2;

	if ((uint32) result2 != result2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));
	result = result2;

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint4int1div);
Datum
uint4int1div(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int8 arg2 = PG_GETARG_INT8(1);
	uint32 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	if (arg2 < 0)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint4int1mod);
Datum
uint4int1mod(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int8 arg2 = PG_GETARG_INT8(1);
	uint32 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	if (arg2 == -1)
		PG_RETURN_UINT32(0);

	result = arg1 % arg2;

	if (arg2 < 0)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint4_to_int1);
Datum
uint4_to_int1(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int8 result;

	result = arg1;
	if ((uint32) result != arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("int1 out of range")));
	if (!SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("int1 out of range")));

	PG_RETURN_INT8(result);
}

PG_FUNCTION_INFO_V1(uint4uint1lt);
Datum
uint4uint1lt(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint4uint1le);
Datum
uint4uint1le(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint4uint1eq);
Datum
uint4uint1eq(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint4uint1ne);
Datum
uint4uint1ne(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint4uint1ge);
Datum
uint4uint1ge(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint4uint1gt);
Datum
uint4uint1gt(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btuint4uint1cmp);
Datum
btuint4uint1cmp(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint4uint1pl);
Datum
uint4uint1pl(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	uint32 result;

	result = arg1 + arg2;

	if (result < arg1 || result < arg2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint4uint1mi);
Datum
uint4uint1mi(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	uint32 result;

	result = arg1 - arg2;

	if (result > arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint4uint1mul);
Datum
uint4uint1mul(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	uint32 result;

	uint64 result2;

	result2 = (uint64)arg1 * (uint64)arg2;

	if ((uint32) result2 != result2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));
	result = result2;

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint4uint1div);
Datum
uint4uint1div(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	uint32 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint4uint1mod);
Datum
uint4uint1mod(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	uint32 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 % arg2;

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint4_to_uint1);
Datum
uint4_to_uint1(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint8 result;

	result = arg1;
	if ((uint32) result != arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("uint1 out of range")));

	PG_RETURN_UINT8(result);
}

PG_FUNCTION_INFO_V1(uint4uint2lt);
Datum
uint4uint2lt(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint4uint2le);
Datum
uint4uint2le(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint4uint2eq);
Datum
uint4uint2eq(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint4uint2ne);
Datum
uint4uint2ne(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint4uint2ge);
Datum
uint4uint2ge(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint4uint2gt);
Datum
uint4uint2gt(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btuint4uint2cmp);
Datum
btuint4uint2cmp(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint4uint2pl);
Datum
uint4uint2pl(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	uint32 result;

	result = arg1 + arg2;

	if (result < arg1 || result < arg2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint4uint2mi);
Datum
uint4uint2mi(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	uint32 result;

	result = arg1 - arg2;

	if (result > arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint4uint2mul);
Datum
uint4uint2mul(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	uint32 result;

	uint64 result2;

	result2 = (uint64)arg1 * (uint64)arg2;

	if ((uint32) result2 != result2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));
	result = result2;

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint4uint2div);
Datum
uint4uint2div(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	uint32 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint4uint2mod);
Datum
uint4uint2mod(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	uint32 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 % arg2;

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint4_to_uint2);
Datum
uint4_to_uint2(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint16 result;

	result = arg1;
	if ((uint32) result != arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("uint2 out of range")));

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(uint4uint4lt);
Datum
uint4uint4lt(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint4uint4le);
Datum
uint4uint4le(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint4uint4eq);
Datum
uint4uint4eq(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint4uint4ne);
Datum
uint4uint4ne(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint4uint4ge);
Datum
uint4uint4ge(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint4uint4gt);
Datum
uint4uint4gt(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btuint4uint4cmp);
Datum
btuint4uint4cmp(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint4uint4pl);
Datum
uint4uint4pl(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	uint32 result;

	result = arg1 + arg2;

	if (result < arg1 || result < arg2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint4uint4mi);
Datum
uint4uint4mi(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	uint32 result;

	result = arg1 - arg2;

	if (result > arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint4uint4mul);
Datum
uint4uint4mul(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	uint32 result;

	uint64 result2;

	result2 = (uint64)arg1 * (uint64)arg2;

	if ((uint32) result2 != result2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));
	result = result2;

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint4uint4div);
Datum
uint4uint4div(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	uint32 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint4uint4mod);
Datum
uint4uint4mod(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	uint32 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 % arg2;

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint4uint8lt);
Datum
uint4uint8lt(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint4uint8le);
Datum
uint4uint8le(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint4uint8eq);
Datum
uint4uint8eq(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint4uint8ne);
Datum
uint4uint8ne(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint4uint8ge);
Datum
uint4uint8ge(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint4uint8gt);
Datum
uint4uint8gt(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btuint4uint8cmp);
Datum
btuint4uint8cmp(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint4uint8pl);
Datum
uint4uint8pl(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	result = arg1 + arg2;

	if (result < arg1 || result < arg2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint4uint8mi);
Datum
uint4uint8mi(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	result = arg1 - arg2;

	if (result > arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint4uint8mul);
Datum
uint4uint8mul(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	result = arg1 * arg2;

	if (((arg1 != ((uint32) arg1) || arg2 != ((uint32) arg2)) && (arg2 != 0 && (result / arg2 != arg1))))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint4uint8div);
Datum
uint4uint8div(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint4uint8mod);
Datum
uint4uint8mod(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 % arg2;

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint4_to_uint8);
Datum
uint4_to_uint8(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint64 result;

	result = arg1;

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint4int2lt);
Datum
uint4int2lt(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int16 arg2 = PG_GETARG_INT16(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint4int2le);
Datum
uint4int2le(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int16 arg2 = PG_GETARG_INT16(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint4int2eq);
Datum
uint4int2eq(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int16 arg2 = PG_GETARG_INT16(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint4int2ne);
Datum
uint4int2ne(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int16 arg2 = PG_GETARG_INT16(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint4int2ge);
Datum
uint4int2ge(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int16 arg2 = PG_GETARG_INT16(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint4int2gt);
Datum
uint4int2gt(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int16 arg2 = PG_GETARG_INT16(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btuint4int2cmp);
Datum
btuint4int2cmp(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int16 arg2 = PG_GETARG_INT16(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint4int2pl);
Datum
uint4int2pl(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int16 arg2 = PG_GETARG_INT16(1);
	uint32 result;

	result = arg1 + arg2;

	if ((arg2 < 0 && result > arg1) || (arg2 > 0 && result < arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint4int2mi);
Datum
uint4int2mi(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int16 arg2 = PG_GETARG_INT16(1);
	uint32 result;

	result = arg1 - arg2;

	if ((arg2 < 0 && result < arg1) || (arg2 > 0 && result > arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint4int2mul);
Datum
uint4int2mul(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int16 arg2 = PG_GETARG_INT16(1);
	uint32 result;

	uint64 result2;

	result2 = (uint64)arg1 * (uint64)arg2;

	if ((uint32) result2 != result2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));
	result = result2;

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint4int2div);
Datum
uint4int2div(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int16 arg2 = PG_GETARG_INT16(1);
	uint32 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	if (arg2 < 0)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint4int2mod);
Datum
uint4int2mod(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int16 arg2 = PG_GETARG_INT16(1);
	uint32 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	if (arg2 == -1)
		PG_RETURN_UINT32(0);

	result = arg1 % arg2;

	if (arg2 < 0)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint4_to_int2);
Datum
uint4_to_int2(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int16 result;

	result = arg1;
	if ((uint32) result != arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("int2 out of range")));
	if (!SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("int2 out of range")));

	PG_RETURN_INT16(result);
}

PG_FUNCTION_INFO_V1(uint4int4lt);
Datum
uint4int4lt(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int32 arg2 = PG_GETARG_INT32(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint4int4le);
Datum
uint4int4le(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int32 arg2 = PG_GETARG_INT32(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint4int4eq);
Datum
uint4int4eq(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int32 arg2 = PG_GETARG_INT32(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint4int4ne);
Datum
uint4int4ne(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int32 arg2 = PG_GETARG_INT32(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint4int4ge);
Datum
uint4int4ge(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int32 arg2 = PG_GETARG_INT32(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint4int4gt);
Datum
uint4int4gt(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int32 arg2 = PG_GETARG_INT32(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btuint4int4cmp);
Datum
btuint4int4cmp(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int32 arg2 = PG_GETARG_INT32(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint4int4pl);
Datum
uint4int4pl(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int32 arg2 = PG_GETARG_INT32(1);
	uint32 result;

	result = arg1 + arg2;

	if ((arg2 < 0 && result > arg1) || (arg2 > 0 && result < arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint4int4mi);
Datum
uint4int4mi(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int32 arg2 = PG_GETARG_INT32(1);
	uint32 result;

	result = arg1 - arg2;

	if ((arg2 < 0 && result < arg1) || (arg2 > 0 && result > arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint4int4mul);
Datum
uint4int4mul(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int32 arg2 = PG_GETARG_INT32(1);
	uint32 result;

	uint64 result2;

	result2 = (uint64)arg1 * (uint64)arg2;

	if ((uint32) result2 != result2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));
	result = result2;

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint4int4div);
Datum
uint4int4div(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int32 arg2 = PG_GETARG_INT32(1);
	uint32 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	if (arg2 < 0)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint4int4mod);
Datum
uint4int4mod(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int32 arg2 = PG_GETARG_INT32(1);
	uint32 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	if (arg2 == -1)
		PG_RETURN_UINT32(0);

	result = arg1 % arg2;

	if (arg2 < 0)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint4_to_int4);
Datum
uint4_to_int4(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int32 result;

	result = arg1;
	if ((uint32) result != arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("int4 out of range")));
	if (!SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("int4 out of range")));

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint4int8lt);
Datum
uint4int8lt(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int64 arg2 = PG_GETARG_INT64(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint4int8le);
Datum
uint4int8le(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int64 arg2 = PG_GETARG_INT64(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint4int8eq);
Datum
uint4int8eq(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int64 arg2 = PG_GETARG_INT64(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint4int8ne);
Datum
uint4int8ne(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int64 arg2 = PG_GETARG_INT64(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint4int8ge);
Datum
uint4int8ge(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int64 arg2 = PG_GETARG_INT64(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint4int8gt);
Datum
uint4int8gt(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int64 arg2 = PG_GETARG_INT64(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btuint4int8cmp);
Datum
btuint4int8cmp(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int64 arg2 = PG_GETARG_INT64(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint4int8pl);
Datum
uint4int8pl(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int64 arg2 = PG_GETARG_INT64(1);
	int64 result;

	result = arg1 + arg2;

	if (SAMESIGN(arg1, arg2) && !SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(uint4int8mi);
Datum
uint4int8mi(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int64 arg2 = PG_GETARG_INT64(1);
	int64 result;

	result = arg1 - arg2;

	if (!SAMESIGN(arg1, arg2) && !SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(uint4int8mul);
Datum
uint4int8mul(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int64 arg2 = PG_GETARG_INT64(1);
	int64 result;

	result = arg1 * arg2;

	if ((arg1 != ((int32) arg1) || arg2 != ((int32) arg2)) && (arg2 != 0 && (result / arg2 != arg1)))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(uint4int8div);
Datum
uint4int8div(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int64 arg2 = PG_GETARG_INT64(1);
	int64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(uint4int8mod);
Datum
uint4int8mod(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int64 arg2 = PG_GETARG_INT64(1);
	int64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	if (arg2 == -1)
		PG_RETURN_INT64(0);

	result = arg1 % arg2;

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(uint4_to_int8);
Datum
uint4_to_int8(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int64 result;

	result = arg1;
	if (!SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("int8 out of range")));

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(uint8int1lt);
Datum
uint8int1lt(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint8int1le);
Datum
uint8int1le(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint8int1eq);
Datum
uint8int1eq(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint8int1ne);
Datum
uint8int1ne(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint8int1ge);
Datum
uint8int1ge(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint8int1gt);
Datum
uint8int1gt(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btuint8int1cmp);
Datum
btuint8int1cmp(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int8 arg2 = PG_GETARG_INT8(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint8int1pl);
Datum
uint8int1pl(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int8 arg2 = PG_GETARG_INT8(1);
	uint64 result;

	result = arg1 + arg2;

	if ((arg2 < 0 && result > arg1) || (arg2 > 0 && result < arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8int1mi);
Datum
uint8int1mi(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int8 arg2 = PG_GETARG_INT8(1);
	uint64 result;

	result = arg1 - arg2;

	if ((arg2 < 0 && result < arg1) || (arg2 > 0 && result > arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8int1mul);
Datum
uint8int1mul(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int8 arg2 = PG_GETARG_INT8(1);
	uint64 result;

	result = arg1 * arg2;

	if ((arg2 < 0) || ((arg1 != ((uint32) arg1) || arg2 != ((uint32) arg2)) && (arg2 != 0 && (result / arg2 != arg1))))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8int1div);
Datum
uint8int1div(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int8 arg2 = PG_GETARG_INT8(1);
	uint64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	if (arg2 < 0)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8int1mod);
Datum
uint8int1mod(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int8 arg2 = PG_GETARG_INT8(1);
	uint64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	if (arg2 == -1)
		PG_RETURN_UINT64(0);

	result = arg1 % arg2;

	if (arg2 < 0)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8_to_int1);
Datum
uint8_to_int1(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int8 result;

	result = arg1;
	if ((uint64) result != arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("int1 out of range")));
	if (!SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("int1 out of range")));

	PG_RETURN_INT8(result);
}

PG_FUNCTION_INFO_V1(uint8uint1lt);
Datum
uint8uint1lt(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint8uint1le);
Datum
uint8uint1le(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint8uint1eq);
Datum
uint8uint1eq(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint8uint1ne);
Datum
uint8uint1ne(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint8uint1ge);
Datum
uint8uint1ge(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint8uint1gt);
Datum
uint8uint1gt(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btuint8uint1cmp);
Datum
btuint8uint1cmp(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint8uint1pl);
Datum
uint8uint1pl(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	uint64 result;

	result = arg1 + arg2;

	if (result < arg1 || result < arg2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8uint1mi);
Datum
uint8uint1mi(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	uint64 result;

	result = arg1 - arg2;

	if (result > arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8uint1mul);
Datum
uint8uint1mul(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	uint64 result;

	result = arg1 * arg2;

	if (((arg1 != ((uint32) arg1) || arg2 != ((uint32) arg2)) && (arg2 != 0 && (result / arg2 != arg1))))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8uint1div);
Datum
uint8uint1div(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	uint64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8uint1mod);
Datum
uint8uint1mod(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	uint64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 % arg2;

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8_to_uint1);
Datum
uint8_to_uint1(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint8 result;

	result = arg1;
	if ((uint64) result != arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("uint1 out of range")));

	PG_RETURN_UINT8(result);
}

PG_FUNCTION_INFO_V1(uint8uint2lt);
Datum
uint8uint2lt(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint8uint2le);
Datum
uint8uint2le(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint8uint2eq);
Datum
uint8uint2eq(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint8uint2ne);
Datum
uint8uint2ne(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint8uint2ge);
Datum
uint8uint2ge(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint8uint2gt);
Datum
uint8uint2gt(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btuint8uint2cmp);
Datum
btuint8uint2cmp(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint8uint2pl);
Datum
uint8uint2pl(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	uint64 result;

	result = arg1 + arg2;

	if (result < arg1 || result < arg2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8uint2mi);
Datum
uint8uint2mi(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	uint64 result;

	result = arg1 - arg2;

	if (result > arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8uint2mul);
Datum
uint8uint2mul(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	uint64 result;

	result = arg1 * arg2;

	if (((arg1 != ((uint32) arg1) || arg2 != ((uint32) arg2)) && (arg2 != 0 && (result / arg2 != arg1))))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8uint2div);
Datum
uint8uint2div(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	uint64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8uint2mod);
Datum
uint8uint2mod(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	uint64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 % arg2;

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8_to_uint2);
Datum
uint8_to_uint2(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint16 result;

	result = arg1;
	if ((uint64) result != arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("uint2 out of range")));

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(uint8uint4lt);
Datum
uint8uint4lt(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint8uint4le);
Datum
uint8uint4le(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint8uint4eq);
Datum
uint8uint4eq(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint8uint4ne);
Datum
uint8uint4ne(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint8uint4ge);
Datum
uint8uint4ge(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint8uint4gt);
Datum
uint8uint4gt(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btuint8uint4cmp);
Datum
btuint8uint4cmp(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint8uint4pl);
Datum
uint8uint4pl(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	uint64 result;

	result = arg1 + arg2;

	if (result < arg1 || result < arg2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8uint4mi);
Datum
uint8uint4mi(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	uint64 result;

	result = arg1 - arg2;

	if (result > arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8uint4mul);
Datum
uint8uint4mul(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	uint64 result;

	result = arg1 * arg2;

	if (((arg1 != ((uint32) arg1) || arg2 != ((uint32) arg2)) && (arg2 != 0 && (result / arg2 != arg1))))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8uint4div);
Datum
uint8uint4div(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	uint64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8uint4mod);
Datum
uint8uint4mod(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	uint64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 % arg2;

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8_to_uint4);
Datum
uint8_to_uint4(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint32 result;

	result = arg1;
	if ((uint64) result != arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("uint4 out of range")));

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint8uint8lt);
Datum
uint8uint8lt(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint8uint8le);
Datum
uint8uint8le(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint8uint8eq);
Datum
uint8uint8eq(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint8uint8ne);
Datum
uint8uint8ne(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint8uint8ge);
Datum
uint8uint8ge(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint8uint8gt);
Datum
uint8uint8gt(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btuint8uint8cmp);
Datum
btuint8uint8cmp(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint8uint8pl);
Datum
uint8uint8pl(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	result = arg1 + arg2;

	if (result < arg1 || result < arg2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8uint8mi);
Datum
uint8uint8mi(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	result = arg1 - arg2;

	if (result > arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8uint8mul);
Datum
uint8uint8mul(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	result = arg1 * arg2;

	if (((arg1 != ((uint32) arg1) || arg2 != ((uint32) arg2)) && (arg2 != 0 && (result / arg2 != arg1))))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8uint8div);
Datum
uint8uint8div(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8uint8mod);
Datum
uint8uint8mod(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 % arg2;

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8int2lt);
Datum
uint8int2lt(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int16 arg2 = PG_GETARG_INT16(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint8int2le);
Datum
uint8int2le(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int16 arg2 = PG_GETARG_INT16(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint8int2eq);
Datum
uint8int2eq(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int16 arg2 = PG_GETARG_INT16(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint8int2ne);
Datum
uint8int2ne(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int16 arg2 = PG_GETARG_INT16(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint8int2ge);
Datum
uint8int2ge(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int16 arg2 = PG_GETARG_INT16(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint8int2gt);
Datum
uint8int2gt(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int16 arg2 = PG_GETARG_INT16(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btuint8int2cmp);
Datum
btuint8int2cmp(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int16 arg2 = PG_GETARG_INT16(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint8int2pl);
Datum
uint8int2pl(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int16 arg2 = PG_GETARG_INT16(1);
	uint64 result;

	result = arg1 + arg2;

	if ((arg2 < 0 && result > arg1) || (arg2 > 0 && result < arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8int2mi);
Datum
uint8int2mi(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int16 arg2 = PG_GETARG_INT16(1);
	uint64 result;

	result = arg1 - arg2;

	if ((arg2 < 0 && result < arg1) || (arg2 > 0 && result > arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8int2mul);
Datum
uint8int2mul(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int16 arg2 = PG_GETARG_INT16(1);
	uint64 result;

	result = arg1 * arg2;

	if ((arg2 < 0) || ((arg1 != ((uint32) arg1) || arg2 != ((uint32) arg2)) && (arg2 != 0 && (result / arg2 != arg1))))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8int2div);
Datum
uint8int2div(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int16 arg2 = PG_GETARG_INT16(1);
	uint64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	if (arg2 < 0)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8int2mod);
Datum
uint8int2mod(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int16 arg2 = PG_GETARG_INT16(1);
	uint64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	if (arg2 == -1)
		PG_RETURN_UINT64(0);

	result = arg1 % arg2;

	if (arg2 < 0)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8_to_int2);
Datum
uint8_to_int2(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int16 result;

	result = arg1;
	if ((uint64) result != arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("int2 out of range")));
	if (!SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("int2 out of range")));

	PG_RETURN_INT16(result);
}

PG_FUNCTION_INFO_V1(uint8int4lt);
Datum
uint8int4lt(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int32 arg2 = PG_GETARG_INT32(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint8int4le);
Datum
uint8int4le(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int32 arg2 = PG_GETARG_INT32(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint8int4eq);
Datum
uint8int4eq(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int32 arg2 = PG_GETARG_INT32(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint8int4ne);
Datum
uint8int4ne(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int32 arg2 = PG_GETARG_INT32(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint8int4ge);
Datum
uint8int4ge(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int32 arg2 = PG_GETARG_INT32(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint8int4gt);
Datum
uint8int4gt(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int32 arg2 = PG_GETARG_INT32(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btuint8int4cmp);
Datum
btuint8int4cmp(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int32 arg2 = PG_GETARG_INT32(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint8int4pl);
Datum
uint8int4pl(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int32 arg2 = PG_GETARG_INT32(1);
	uint64 result;

	result = arg1 + arg2;

	if ((arg2 < 0 && result > arg1) || (arg2 > 0 && result < arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8int4mi);
Datum
uint8int4mi(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int32 arg2 = PG_GETARG_INT32(1);
	uint64 result;

	result = arg1 - arg2;

	if ((arg2 < 0 && result < arg1) || (arg2 > 0 && result > arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8int4mul);
Datum
uint8int4mul(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int32 arg2 = PG_GETARG_INT32(1);
	uint64 result;

	result = arg1 * arg2;

	if ((arg2 < 0) || ((arg1 != ((uint32) arg1) || arg2 != ((uint32) arg2)) && (arg2 != 0 && (result / arg2 != arg1))))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8int4div);
Datum
uint8int4div(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int32 arg2 = PG_GETARG_INT32(1);
	uint64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	if (arg2 < 0)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8int4mod);
Datum
uint8int4mod(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int32 arg2 = PG_GETARG_INT32(1);
	uint64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	if (arg2 == -1)
		PG_RETURN_UINT64(0);

	result = arg1 % arg2;

	if (arg2 < 0)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8_to_int4);
Datum
uint8_to_int4(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int32 result;

	result = arg1;
	if ((uint64) result != arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("int4 out of range")));
	if (!SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("int4 out of range")));

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint8int8lt);
Datum
uint8int8lt(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int64 arg2 = PG_GETARG_INT64(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint8int8le);
Datum
uint8int8le(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int64 arg2 = PG_GETARG_INT64(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint8int8eq);
Datum
uint8int8eq(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int64 arg2 = PG_GETARG_INT64(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint8int8ne);
Datum
uint8int8ne(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int64 arg2 = PG_GETARG_INT64(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint8int8ge);
Datum
uint8int8ge(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int64 arg2 = PG_GETARG_INT64(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uint8int8gt);
Datum
uint8int8gt(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int64 arg2 = PG_GETARG_INT64(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btuint8int8cmp);
Datum
btuint8int8cmp(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int64 arg2 = PG_GETARG_INT64(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uint8int8pl);
Datum
uint8int8pl(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int64 arg2 = PG_GETARG_INT64(1);
	uint64 result;

	result = arg1 + arg2;

	if ((arg2 < 0 && result > arg1) || (arg2 > 0 && result < arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8int8mi);
Datum
uint8int8mi(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int64 arg2 = PG_GETARG_INT64(1);
	uint64 result;

	result = arg1 - arg2;

	if ((arg2 < 0 && result < arg1) || (arg2 > 0 && result > arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8int8mul);
Datum
uint8int8mul(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int64 arg2 = PG_GETARG_INT64(1);
	uint64 result;

	result = arg1 * arg2;

	if ((arg2 < 0) || ((arg1 != ((uint32) arg1) || arg2 != ((uint32) arg2)) && (arg2 != 0 && (result / arg2 != arg1))))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8int8div);
Datum
uint8int8div(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int64 arg2 = PG_GETARG_INT64(1);
	uint64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	if (arg2 < 0)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8int8mod);
Datum
uint8int8mod(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int64 arg2 = PG_GETARG_INT64(1);
	uint64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	if (arg2 == -1)
		PG_RETURN_UINT64(0);

	result = arg1 % arg2;

	if (arg2 < 0)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8_to_int8);
Datum
uint8_to_int8(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int64 result;

	result = arg1;
	if ((uint64) result != arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("int8 out of range")));
	if (!SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("int8 out of range")));

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(int2int1lt);
Datum
int2int1lt(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int2int1le);
Datum
int2int1le(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int2int1eq);
Datum
int2int1eq(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int2int1ne);
Datum
int2int1ne(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int2int1ge);
Datum
int2int1ge(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int2int1gt);
Datum
int2int1gt(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btint2int1cmp);
Datum
btint2int1cmp(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	int8 arg2 = PG_GETARG_INT8(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int2int1pl);
Datum
int2int1pl(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	int8 arg2 = PG_GETARG_INT8(1);
	int16 result;

	result = arg1 + arg2;

	if (SAMESIGN(arg1, arg2) && !SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT16(result);
}

PG_FUNCTION_INFO_V1(int2int1mi);
Datum
int2int1mi(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	int8 arg2 = PG_GETARG_INT8(1);
	int16 result;

	result = arg1 - arg2;

	if (!SAMESIGN(arg1, arg2) && !SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT16(result);
}

PG_FUNCTION_INFO_V1(int2int1mul);
Datum
int2int1mul(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	int8 arg2 = PG_GETARG_INT8(1);
	int16 result;

	int32 result2;

	result2 = (int32)arg1 * (int32)arg2;

	if ((int16) result2 != result2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));
	result = result2;

	PG_RETURN_INT16(result);
}

PG_FUNCTION_INFO_V1(int2int1div);
Datum
int2int1div(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	int8 arg2 = PG_GETARG_INT8(1);
	int16 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	PG_RETURN_INT16(result);
}

PG_FUNCTION_INFO_V1(int2int1mod);
Datum
int2int1mod(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	int8 arg2 = PG_GETARG_INT8(1);
	int16 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	if (arg2 == -1)
		PG_RETURN_INT16(0);

	result = arg1 % arg2;

	PG_RETURN_INT16(result);
}

PG_FUNCTION_INFO_V1(int2_to_int1);
Datum
int2_to_int1(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	int8 result;

	result = arg1;
	if ((int16) result != arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("int1 out of range")));

	PG_RETURN_INT8(result);
}

PG_FUNCTION_INFO_V1(int2uint1lt);
Datum
int2uint1lt(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int2uint1le);
Datum
int2uint1le(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int2uint1eq);
Datum
int2uint1eq(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int2uint1ne);
Datum
int2uint1ne(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int2uint1ge);
Datum
int2uint1ge(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int2uint1gt);
Datum
int2uint1gt(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btint2uint1cmp);
Datum
btint2uint1cmp(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int2uint1pl);
Datum
int2uint1pl(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	int16 result;

	result = arg1 + arg2;

	if (SAMESIGN(arg1, arg2) && !SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT16(result);
}

PG_FUNCTION_INFO_V1(int2uint1mi);
Datum
int2uint1mi(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	int16 result;

	result = arg1 - arg2;

	if (!SAMESIGN(arg1, arg2) && !SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT16(result);
}

PG_FUNCTION_INFO_V1(int2uint1mul);
Datum
int2uint1mul(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	int16 result;

	int32 result2;

	result2 = (int32)arg1 * (int32)arg2;

	if ((int16) result2 != result2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));
	result = result2;

	PG_RETURN_INT16(result);
}

PG_FUNCTION_INFO_V1(int2uint1div);
Datum
int2uint1div(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	int16 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	PG_RETURN_INT16(result);
}

PG_FUNCTION_INFO_V1(int2uint1mod);
Datum
int2uint1mod(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	int16 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 % arg2;

	PG_RETURN_INT16(result);
}

PG_FUNCTION_INFO_V1(int2_to_uint1);
Datum
int2_to_uint1(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint8 result;

	result = arg1;
	if ((int16) result != arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("uint1 out of range")));
	if (!SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("uint1 out of range")));

	PG_RETURN_UINT8(result);
}

PG_FUNCTION_INFO_V1(int2uint2lt);
Datum
int2uint2lt(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int2uint2le);
Datum
int2uint2le(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int2uint2eq);
Datum
int2uint2eq(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int2uint2ne);
Datum
int2uint2ne(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int2uint2ge);
Datum
int2uint2ge(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int2uint2gt);
Datum
int2uint2gt(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btint2uint2cmp);
Datum
btint2uint2cmp(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int2uint2pl);
Datum
int2uint2pl(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	uint16 result;

	result = arg1 + arg2;

	if ((arg1 < 0 && result > arg2) || (arg1 > 0 && result < arg2))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(int2uint2mi);
Datum
int2uint2mi(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	uint16 result;

	result = arg1 - arg2;

	if ((arg1 < 0) || (result > arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(int2uint2mul);
Datum
int2uint2mul(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	uint16 result;

	uint32 result2;

	result2 = (uint32)arg1 * (uint32)arg2;

	if ((uint16) result2 != result2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));
	result = result2;

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(int2uint2div);
Datum
int2uint2div(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	uint16 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	if (arg1 < 0)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(int2uint2mod);
Datum
int2uint2mod(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	uint16 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 % arg2;

	if (arg1 < 0)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(int2_to_uint2);
Datum
int2_to_uint2(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint16 result;

	result = arg1;
	if ((int16) result != arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("uint2 out of range")));
	if (!SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("uint2 out of range")));

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(int2uint4lt);
Datum
int2uint4lt(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int2uint4le);
Datum
int2uint4le(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int2uint4eq);
Datum
int2uint4eq(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int2uint4ne);
Datum
int2uint4ne(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int2uint4ge);
Datum
int2uint4ge(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int2uint4gt);
Datum
int2uint4gt(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btint2uint4cmp);
Datum
btint2uint4cmp(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int2uint4pl);
Datum
int2uint4pl(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	uint32 result;

	result = arg1 + arg2;

	if ((arg1 < 0 && result > arg2) || (arg1 > 0 && result < arg2))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(int2uint4mi);
Datum
int2uint4mi(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	uint32 result;

	result = arg1 - arg2;

	if ((arg1 < 0) || (result > arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(int2uint4mul);
Datum
int2uint4mul(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	uint32 result;

	uint64 result2;

	result2 = (uint64)arg1 * (uint64)arg2;

	if ((uint32) result2 != result2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));
	result = result2;

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(int2uint4div);
Datum
int2uint4div(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	uint32 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	if (arg1 < 0)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(int2uint4mod);
Datum
int2uint4mod(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	uint32 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 % arg2;

	if (arg1 < 0)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(int2_to_uint4);
Datum
int2_to_uint4(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint32 result;

	result = arg1;
	if (!SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("uint4 out of range")));

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(int2uint8lt);
Datum
int2uint8lt(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int2uint8le);
Datum
int2uint8le(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int2uint8eq);
Datum
int2uint8eq(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int2uint8ne);
Datum
int2uint8ne(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int2uint8ge);
Datum
int2uint8ge(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int2uint8gt);
Datum
int2uint8gt(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btint2uint8cmp);
Datum
btint2uint8cmp(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int2uint8pl);
Datum
int2uint8pl(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	result = arg1 + arg2;

	if ((arg1 < 0 && result > arg2) || (arg1 > 0 && result < arg2))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(int2uint8mi);
Datum
int2uint8mi(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	result = arg1 - arg2;

	if ((arg1 < 0) || (result > arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(int2uint8mul);
Datum
int2uint8mul(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	result = arg1 * arg2;

	if ((arg1 < 0) || ((arg1 != ((uint32) arg1) || arg2 != ((uint32) arg2)) && (arg2 != 0 && (result / arg2 != arg1))))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(int2uint8div);
Datum
int2uint8div(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	if (arg1 < 0)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(int2uint8mod);
Datum
int2uint8mod(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 % arg2;

	if (arg1 < 0)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(int2_to_uint8);
Datum
int2_to_uint8(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	uint64 result;

	result = arg1;
	if (!SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("uint8 out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(int4int1lt);
Datum
int4int1lt(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int4int1le);
Datum
int4int1le(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int4int1eq);
Datum
int4int1eq(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int4int1ne);
Datum
int4int1ne(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int4int1ge);
Datum
int4int1ge(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int4int1gt);
Datum
int4int1gt(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btint4int1cmp);
Datum
btint4int1cmp(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	int8 arg2 = PG_GETARG_INT8(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int4int1pl);
Datum
int4int1pl(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	int8 arg2 = PG_GETARG_INT8(1);
	int32 result;

	result = arg1 + arg2;

	if (SAMESIGN(arg1, arg2) && !SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int4int1mi);
Datum
int4int1mi(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	int8 arg2 = PG_GETARG_INT8(1);
	int32 result;

	result = arg1 - arg2;

	if (!SAMESIGN(arg1, arg2) && !SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int4int1mul);
Datum
int4int1mul(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	int8 arg2 = PG_GETARG_INT8(1);
	int32 result;

	int64 result2;

	result2 = (int64)arg1 * (int64)arg2;

	if ((int32) result2 != result2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));
	result = result2;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int4int1div);
Datum
int4int1div(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	int8 arg2 = PG_GETARG_INT8(1);
	int32 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int4int1mod);
Datum
int4int1mod(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	int8 arg2 = PG_GETARG_INT8(1);
	int32 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	if (arg2 == -1)
		PG_RETURN_INT32(0);

	result = arg1 % arg2;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int4_to_int1);
Datum
int4_to_int1(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	int8 result;

	result = arg1;
	if ((int32) result != arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("int1 out of range")));

	PG_RETURN_INT8(result);
}

PG_FUNCTION_INFO_V1(int4uint1lt);
Datum
int4uint1lt(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int4uint1le);
Datum
int4uint1le(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int4uint1eq);
Datum
int4uint1eq(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int4uint1ne);
Datum
int4uint1ne(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int4uint1ge);
Datum
int4uint1ge(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int4uint1gt);
Datum
int4uint1gt(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btint4uint1cmp);
Datum
btint4uint1cmp(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int4uint1pl);
Datum
int4uint1pl(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	int32 result;

	result = arg1 + arg2;

	if (SAMESIGN(arg1, arg2) && !SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int4uint1mi);
Datum
int4uint1mi(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	int32 result;

	result = arg1 - arg2;

	if (!SAMESIGN(arg1, arg2) && !SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int4uint1mul);
Datum
int4uint1mul(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	int32 result;

	int64 result2;

	result2 = (int64)arg1 * (int64)arg2;

	if ((int32) result2 != result2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));
	result = result2;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int4uint1div);
Datum
int4uint1div(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	int32 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int4uint1mod);
Datum
int4uint1mod(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	int32 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 % arg2;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int4_to_uint1);
Datum
int4_to_uint1(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint8 result;

	result = arg1;
	if ((int32) result != arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("uint1 out of range")));
	if (!SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("uint1 out of range")));

	PG_RETURN_UINT8(result);
}

PG_FUNCTION_INFO_V1(int4uint2lt);
Datum
int4uint2lt(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int4uint2le);
Datum
int4uint2le(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int4uint2eq);
Datum
int4uint2eq(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int4uint2ne);
Datum
int4uint2ne(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int4uint2ge);
Datum
int4uint2ge(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int4uint2gt);
Datum
int4uint2gt(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btint4uint2cmp);
Datum
btint4uint2cmp(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int4uint2pl);
Datum
int4uint2pl(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	int32 result;

	result = arg1 + arg2;

	if (SAMESIGN(arg1, arg2) && !SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int4uint2mi);
Datum
int4uint2mi(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	int32 result;

	result = arg1 - arg2;

	if (!SAMESIGN(arg1, arg2) && !SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int4uint2mul);
Datum
int4uint2mul(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	int32 result;

	int64 result2;

	result2 = (int64)arg1 * (int64)arg2;

	if ((int32) result2 != result2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));
	result = result2;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int4uint2div);
Datum
int4uint2div(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	int32 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int4uint2mod);
Datum
int4uint2mod(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	int32 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 % arg2;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int4_to_uint2);
Datum
int4_to_uint2(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint16 result;

	result = arg1;
	if ((int32) result != arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("uint2 out of range")));
	if (!SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("uint2 out of range")));

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(int4uint4lt);
Datum
int4uint4lt(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int4uint4le);
Datum
int4uint4le(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int4uint4eq);
Datum
int4uint4eq(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int4uint4ne);
Datum
int4uint4ne(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int4uint4ge);
Datum
int4uint4ge(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int4uint4gt);
Datum
int4uint4gt(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btint4uint4cmp);
Datum
btint4uint4cmp(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int4uint4pl);
Datum
int4uint4pl(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	uint32 result;

	result = arg1 + arg2;

	if ((arg1 < 0 && result > arg2) || (arg1 > 0 && result < arg2))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(int4uint4mi);
Datum
int4uint4mi(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	uint32 result;

	result = arg1 - arg2;

	if ((arg1 < 0) || (result > arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(int4uint4mul);
Datum
int4uint4mul(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	uint32 result;

	uint64 result2;

	result2 = (uint64)arg1 * (uint64)arg2;

	if ((uint32) result2 != result2)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));
	result = result2;

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(int4uint4div);
Datum
int4uint4div(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	uint32 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	if (arg1 < 0)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(int4uint4mod);
Datum
int4uint4mod(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	uint32 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 % arg2;

	if (arg1 < 0)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(int4_to_uint4);
Datum
int4_to_uint4(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint32 result;

	result = arg1;
	if ((int32) result != arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("uint4 out of range")));
	if (!SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("uint4 out of range")));

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(int4uint8lt);
Datum
int4uint8lt(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int4uint8le);
Datum
int4uint8le(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int4uint8eq);
Datum
int4uint8eq(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int4uint8ne);
Datum
int4uint8ne(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int4uint8ge);
Datum
int4uint8ge(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int4uint8gt);
Datum
int4uint8gt(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btint4uint8cmp);
Datum
btint4uint8cmp(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int4uint8pl);
Datum
int4uint8pl(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	result = arg1 + arg2;

	if ((arg1 < 0 && result > arg2) || (arg1 > 0 && result < arg2))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(int4uint8mi);
Datum
int4uint8mi(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	result = arg1 - arg2;

	if ((arg1 < 0) || (result > arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(int4uint8mul);
Datum
int4uint8mul(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	result = arg1 * arg2;

	if ((arg1 < 0) || ((arg1 != ((uint32) arg1) || arg2 != ((uint32) arg2)) && (arg2 != 0 && (result / arg2 != arg1))))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(int4uint8div);
Datum
int4uint8div(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	if (arg1 < 0)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(int4uint8mod);
Datum
int4uint8mod(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 % arg2;

	if (arg1 < 0)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(int4_to_uint8);
Datum
int4_to_uint8(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	uint64 result;

	result = arg1;
	if (!SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("uint8 out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(int8int1lt);
Datum
int8int1lt(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int8int1le);
Datum
int8int1le(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int8int1eq);
Datum
int8int1eq(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int8int1ne);
Datum
int8int1ne(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int8int1ge);
Datum
int8int1ge(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int8int1gt);
Datum
int8int1gt(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	int8 arg2 = PG_GETARG_INT8(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btint8int1cmp);
Datum
btint8int1cmp(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	int8 arg2 = PG_GETARG_INT8(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int8int1pl);
Datum
int8int1pl(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	int8 arg2 = PG_GETARG_INT8(1);
	int64 result;

	result = arg1 + arg2;

	if (SAMESIGN(arg1, arg2) && !SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(int8int1mi);
Datum
int8int1mi(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	int8 arg2 = PG_GETARG_INT8(1);
	int64 result;

	result = arg1 - arg2;

	if (!SAMESIGN(arg1, arg2) && !SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(int8int1mul);
Datum
int8int1mul(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	int8 arg2 = PG_GETARG_INT8(1);
	int64 result;

	result = arg1 * arg2;

	if ((arg1 != ((int32) arg1) || arg2 != ((int32) arg2)) && (arg2 != 0 && ((arg2 == -1 && arg1 < 0 && result < 0) || result / arg2 != arg1)))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(int8int1div);
Datum
int8int1div(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	int8 arg2 = PG_GETARG_INT8(1);
	int64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(int8int1mod);
Datum
int8int1mod(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	int8 arg2 = PG_GETARG_INT8(1);
	int64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	if (arg2 == -1)
		PG_RETURN_INT64(0);

	result = arg1 % arg2;

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(int8_to_int1);
Datum
int8_to_int1(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	int8 result;

	result = arg1;
	if ((int64) result != arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("int1 out of range")));

	PG_RETURN_INT8(result);
}

PG_FUNCTION_INFO_V1(int8uint1lt);
Datum
int8uint1lt(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int8uint1le);
Datum
int8uint1le(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int8uint1eq);
Datum
int8uint1eq(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int8uint1ne);
Datum
int8uint1ne(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int8uint1ge);
Datum
int8uint1ge(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int8uint1gt);
Datum
int8uint1gt(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btint8uint1cmp);
Datum
btint8uint1cmp(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int8uint1pl);
Datum
int8uint1pl(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	int64 result;

	result = arg1 + arg2;

	if (SAMESIGN(arg1, arg2) && !SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(int8uint1mi);
Datum
int8uint1mi(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	int64 result;

	result = arg1 - arg2;

	if (!SAMESIGN(arg1, arg2) && !SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(int8uint1mul);
Datum
int8uint1mul(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	int64 result;

	result = arg1 * arg2;

	if ((arg1 != ((int32) arg1) || arg2 != ((int32) arg2)) && (arg2 != 0 && (result / arg2 != arg1)))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(int8uint1div);
Datum
int8uint1div(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	int64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(int8uint1mod);
Datum
int8uint1mod(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	int64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 % arg2;

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(int8_to_uint1);
Datum
int8_to_uint1(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint8 result;

	result = arg1;
	if ((int64) result != arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("uint1 out of range")));
	if (!SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("uint1 out of range")));

	PG_RETURN_UINT8(result);
}

PG_FUNCTION_INFO_V1(int8uint2lt);
Datum
int8uint2lt(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int8uint2le);
Datum
int8uint2le(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int8uint2eq);
Datum
int8uint2eq(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int8uint2ne);
Datum
int8uint2ne(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int8uint2ge);
Datum
int8uint2ge(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int8uint2gt);
Datum
int8uint2gt(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btint8uint2cmp);
Datum
btint8uint2cmp(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int8uint2pl);
Datum
int8uint2pl(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	int64 result;

	result = arg1 + arg2;

	if (SAMESIGN(arg1, arg2) && !SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(int8uint2mi);
Datum
int8uint2mi(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	int64 result;

	result = arg1 - arg2;

	if (!SAMESIGN(arg1, arg2) && !SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(int8uint2mul);
Datum
int8uint2mul(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	int64 result;

	result = arg1 * arg2;

	if ((arg1 != ((int32) arg1) || arg2 != ((int32) arg2)) && (arg2 != 0 && (result / arg2 != arg1)))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(int8uint2div);
Datum
int8uint2div(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	int64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(int8uint2mod);
Datum
int8uint2mod(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	int64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 % arg2;

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(int8_to_uint2);
Datum
int8_to_uint2(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint16 result;

	result = arg1;
	if ((int64) result != arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("uint2 out of range")));
	if (!SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("uint2 out of range")));

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(int8uint4lt);
Datum
int8uint4lt(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int8uint4le);
Datum
int8uint4le(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int8uint4eq);
Datum
int8uint4eq(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int8uint4ne);
Datum
int8uint4ne(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int8uint4ge);
Datum
int8uint4ge(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int8uint4gt);
Datum
int8uint4gt(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btint8uint4cmp);
Datum
btint8uint4cmp(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int8uint4pl);
Datum
int8uint4pl(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	int64 result;

	result = arg1 + arg2;

	if (SAMESIGN(arg1, arg2) && !SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(int8uint4mi);
Datum
int8uint4mi(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	int64 result;

	result = arg1 - arg2;

	if (!SAMESIGN(arg1, arg2) && !SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(int8uint4mul);
Datum
int8uint4mul(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	int64 result;

	result = arg1 * arg2;

	if ((arg1 != ((int32) arg1) || arg2 != ((int32) arg2)) && (arg2 != 0 && (result / arg2 != arg1)))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(int8uint4div);
Datum
int8uint4div(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	int64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(int8uint4mod);
Datum
int8uint4mod(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	int64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 % arg2;

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(int8_to_uint4);
Datum
int8_to_uint4(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint32 result;

	result = arg1;
	if ((int64) result != arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("uint4 out of range")));
	if (!SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("uint4 out of range")));

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(int8uint8lt);
Datum
int8uint8lt(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 < arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int8uint8le);
Datum
int8uint8le(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 <= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int8uint8eq);
Datum
int8uint8eq(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 == arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int8uint8ne);
Datum
int8uint8ne(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 != arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int8uint8ge);
Datum
int8uint8ge(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 >= arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(int8uint8gt);
Datum
int8uint8gt(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	bool result;

	result = arg1 > arg2;

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(btint8uint8cmp);
Datum
btint8uint8cmp(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	int32 result;

	if (arg1 > arg2)
		result = 1;
	else if (arg1 == arg2)
		result = 0;
	else
		result = -1;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int8uint8pl);
Datum
int8uint8pl(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	result = arg1 + arg2;

	if ((arg1 < 0 && result > arg2) || (arg1 > 0 && result < arg2))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(int8uint8mi);
Datum
int8uint8mi(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	result = arg1 - arg2;

	if ((arg1 < 0) || (result > arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(int8uint8mul);
Datum
int8uint8mul(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	result = arg1 * arg2;

	if ((arg1 < 0) || ((arg1 != ((uint32) arg1) || arg2 != ((uint32) arg2)) && (arg2 != 0 && (result / arg2 != arg1))))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(int8uint8div);
Datum
int8uint8div(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 / arg2;

	if (arg1 < 0)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(int8uint8mod);
Datum
int8uint8mod(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	result = arg1 % arg2;

	if (arg1 < 0)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("integer out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(int8_to_uint8);
Datum
int8_to_uint8(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	uint64 result;

	result = arg1;
	if ((int64) result != arg1)
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("uint8 out of range")));
	if (!SAMESIGN(result, arg1))
		ereport(ERROR,
			(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
			 errmsg("uint8 out of range")));

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(int1int1and);
Datum
int1int1and(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int8 arg2 = PG_GETARG_INT8(1);
	int8 result;

	result = arg1 & arg2;

	PG_RETURN_INT8(result);
}

PG_FUNCTION_INFO_V1(int1int1or);
Datum
int1int1or(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int8 arg2 = PG_GETARG_INT8(1);
	int8 result;

	result = arg1 | arg2;

	PG_RETURN_INT8(result);
}

PG_FUNCTION_INFO_V1(int1int1xor);
Datum
int1int1xor(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int8 arg2 = PG_GETARG_INT8(1);
	int8 result;

	result = arg1 ^ arg2;

	PG_RETURN_INT8(result);
}

PG_FUNCTION_INFO_V1(int1not);
Datum
int1not(PG_FUNCTION_ARGS)
{
	int8 arg2 = PG_GETARG_INT8(0);
	int8 result;

	result =  ~ arg2;

	PG_RETURN_INT8(result);
}

PG_FUNCTION_INFO_V1(int1int4shl);
Datum
int1int4shl(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int32 arg2 = PG_GETARG_INT32(1);
	int8 result;

	result = arg1 << arg2;

	PG_RETURN_INT8(result);
}

PG_FUNCTION_INFO_V1(int1int4shr);
Datum
int1int4shr(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int32 arg2 = PG_GETARG_INT32(1);
	int8 result;

	result = arg1 >> arg2;

	PG_RETURN_INT8(result);
}

static int
btint1fastcmp(Datum x, Datum y, SortSupport ssup)
{
	int8 a = DatumGetInt8(x);
	int8 b = DatumGetInt8(y);

	if (a > b)
		return 1;
	else if (a == b)
		return 0;
	else
		return -1;
}

PG_FUNCTION_INFO_V1(btint1sortsupport);
Datum
btint1sortsupport(PG_FUNCTION_ARGS)
{
	SortSupport ssup = (SortSupport) PG_GETARG_POINTER(0);

	ssup->comparator = btint1fastcmp;
	PG_RETURN_VOID();
}

PG_FUNCTION_INFO_V1(int1smaller);
Datum
int1smaller(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int8 arg2 = PG_GETARG_INT8(1);
	int8 result;

	result = (arg1 < arg2) ? arg1 : arg2;

	PG_RETURN_INT8(result);
}

PG_FUNCTION_INFO_V1(int1larger);
Datum
int1larger(PG_FUNCTION_ARGS)
{
	int8 arg1 = PG_GETARG_INT8(0);
	int8 arg2 = PG_GETARG_INT8(1);
	int8 result;

	result = (arg1 > arg2) ? arg1 : arg2;

	PG_RETURN_INT8(result);
}

PG_FUNCTION_INFO_V1(uint1uint1and);
Datum
uint1uint1and(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	uint8 result;

	result = arg1 & arg2;

	PG_RETURN_UINT8(result);
}

PG_FUNCTION_INFO_V1(uint1uint1or);
Datum
uint1uint1or(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	uint8 result;

	result = arg1 | arg2;

	PG_RETURN_UINT8(result);
}

PG_FUNCTION_INFO_V1(uint1uint1xor);
Datum
uint1uint1xor(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	uint8 result;

	result = arg1 ^ arg2;

	PG_RETURN_UINT8(result);
}

PG_FUNCTION_INFO_V1(uint1not);
Datum
uint1not(PG_FUNCTION_ARGS)
{
	uint8 arg2 = PG_GETARG_UINT8(0);
	uint8 result;

	result =  ~ arg2;

	PG_RETURN_UINT8(result);
}

PG_FUNCTION_INFO_V1(uint1int4shl);
Datum
uint1int4shl(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int32 arg2 = PG_GETARG_INT32(1);
	uint8 result;

	result = arg1 << arg2;

	PG_RETURN_UINT8(result);
}

PG_FUNCTION_INFO_V1(uint1int4shr);
Datum
uint1int4shr(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	int32 arg2 = PG_GETARG_INT32(1);
	uint8 result;

	result = arg1 >> arg2;

	PG_RETURN_UINT8(result);
}

static int
btuint1fastcmp(Datum x, Datum y, SortSupport ssup)
{
	uint8 a = DatumGetUInt8(x);
	uint8 b = DatumGetUInt8(y);

	if (a > b)
		return 1;
	else if (a == b)
		return 0;
	else
		return -1;
}

PG_FUNCTION_INFO_V1(btuint1sortsupport);
Datum
btuint1sortsupport(PG_FUNCTION_ARGS)
{
	SortSupport ssup = (SortSupport) PG_GETARG_POINTER(0);

	ssup->comparator = btuint1fastcmp;
	PG_RETURN_VOID();
}

PG_FUNCTION_INFO_V1(uint1smaller);
Datum
uint1smaller(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	uint8 result;

	result = (arg1 < arg2) ? arg1 : arg2;

	PG_RETURN_UINT8(result);
}

PG_FUNCTION_INFO_V1(uint1larger);
Datum
uint1larger(PG_FUNCTION_ARGS)
{
	uint8 arg1 = PG_GETARG_UINT8(0);
	uint8 arg2 = PG_GETARG_UINT8(1);
	uint8 result;

	result = (arg1 > arg2) ? arg1 : arg2;

	PG_RETURN_UINT8(result);
}

PG_FUNCTION_INFO_V1(uint2uint2and);
Datum
uint2uint2and(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	uint16 result;

	result = arg1 & arg2;

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(uint2uint2or);
Datum
uint2uint2or(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	uint16 result;

	result = arg1 | arg2;

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(uint2uint2xor);
Datum
uint2uint2xor(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	uint16 result;

	result = arg1 ^ arg2;

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(uint2not);
Datum
uint2not(PG_FUNCTION_ARGS)
{
	uint16 arg2 = PG_GETARG_UINT16(0);
	uint16 result;

	result =  ~ arg2;

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(uint2int4shl);
Datum
uint2int4shl(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int32 arg2 = PG_GETARG_INT32(1);
	uint16 result;

	result = arg1 << arg2;

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(uint2int4shr);
Datum
uint2int4shr(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	int32 arg2 = PG_GETARG_INT32(1);
	uint16 result;

	result = arg1 >> arg2;

	PG_RETURN_UINT16(result);
}

static int
btuint2fastcmp(Datum x, Datum y, SortSupport ssup)
{
	uint16 a = DatumGetUInt16(x);
	uint16 b = DatumGetUInt16(y);

	if (a > b)
		return 1;
	else if (a == b)
		return 0;
	else
		return -1;
}

PG_FUNCTION_INFO_V1(btuint2sortsupport);
Datum
btuint2sortsupport(PG_FUNCTION_ARGS)
{
	SortSupport ssup = (SortSupport) PG_GETARG_POINTER(0);

	ssup->comparator = btuint2fastcmp;
	PG_RETURN_VOID();
}

PG_FUNCTION_INFO_V1(uint2smaller);
Datum
uint2smaller(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	uint16 result;

	result = (arg1 < arg2) ? arg1 : arg2;

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(uint2larger);
Datum
uint2larger(PG_FUNCTION_ARGS)
{
	uint16 arg1 = PG_GETARG_UINT16(0);
	uint16 arg2 = PG_GETARG_UINT16(1);
	uint16 result;

	result = (arg1 > arg2) ? arg1 : arg2;

	PG_RETURN_UINT16(result);
}

PG_FUNCTION_INFO_V1(uint4uint4and);
Datum
uint4uint4and(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	uint32 result;

	result = arg1 & arg2;

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint4uint4or);
Datum
uint4uint4or(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	uint32 result;

	result = arg1 | arg2;

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint4uint4xor);
Datum
uint4uint4xor(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	uint32 result;

	result = arg1 ^ arg2;

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint4not);
Datum
uint4not(PG_FUNCTION_ARGS)
{
	uint32 arg2 = PG_GETARG_UINT32(0);
	uint32 result;

	result =  ~ arg2;

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint4int4shl);
Datum
uint4int4shl(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int32 arg2 = PG_GETARG_INT32(1);
	uint32 result;

	result = arg1 << arg2;

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint4int4shr);
Datum
uint4int4shr(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	int32 arg2 = PG_GETARG_INT32(1);
	uint32 result;

	result = arg1 >> arg2;

	PG_RETURN_UINT32(result);
}

static int
btuint4fastcmp(Datum x, Datum y, SortSupport ssup)
{
	uint32 a = DatumGetUInt32(x);
	uint32 b = DatumGetUInt32(y);

	if (a > b)
		return 1;
	else if (a == b)
		return 0;
	else
		return -1;
}

PG_FUNCTION_INFO_V1(btuint4sortsupport);
Datum
btuint4sortsupport(PG_FUNCTION_ARGS)
{
	SortSupport ssup = (SortSupport) PG_GETARG_POINTER(0);

	ssup->comparator = btuint4fastcmp;
	PG_RETURN_VOID();
}

PG_FUNCTION_INFO_V1(uint4smaller);
Datum
uint4smaller(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	uint32 result;

	result = (arg1 < arg2) ? arg1 : arg2;

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint4larger);
Datum
uint4larger(PG_FUNCTION_ARGS)
{
	uint32 arg1 = PG_GETARG_UINT32(0);
	uint32 arg2 = PG_GETARG_UINT32(1);
	uint32 result;

	result = (arg1 > arg2) ? arg1 : arg2;

	PG_RETURN_UINT32(result);
}

PG_FUNCTION_INFO_V1(uint8uint8and);
Datum
uint8uint8and(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	result = arg1 & arg2;

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8uint8or);
Datum
uint8uint8or(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	result = arg1 | arg2;

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8uint8xor);
Datum
uint8uint8xor(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	result = arg1 ^ arg2;

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8not);
Datum
uint8not(PG_FUNCTION_ARGS)
{
	uint64 arg2 = PG_GETARG_UINT64(0);
	uint64 result;

	result =  ~ arg2;

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8int4shl);
Datum
uint8int4shl(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int32 arg2 = PG_GETARG_INT32(1);
	uint64 result;

	result = arg1 << arg2;

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8int4shr);
Datum
uint8int4shr(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	int32 arg2 = PG_GETARG_INT32(1);
	uint64 result;

	result = arg1 >> arg2;

	PG_RETURN_UINT64(result);
}

static int
btuint8fastcmp(Datum x, Datum y, SortSupport ssup)
{
	uint64 a = DatumGetUInt64(x);
	uint64 b = DatumGetUInt64(y);

	if (a > b)
		return 1;
	else if (a == b)
		return 0;
	else
		return -1;
}

PG_FUNCTION_INFO_V1(btuint8sortsupport);
Datum
btuint8sortsupport(PG_FUNCTION_ARGS)
{
	SortSupport ssup = (SortSupport) PG_GETARG_POINTER(0);

	ssup->comparator = btuint8fastcmp;
	PG_RETURN_VOID();
}

PG_FUNCTION_INFO_V1(uint8smaller);
Datum
uint8smaller(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	result = (arg1 < arg2) ? arg1 : arg2;

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(uint8larger);
Datum
uint8larger(PG_FUNCTION_ARGS)
{
	uint64 arg1 = PG_GETARG_UINT64(0);
	uint64 arg2 = PG_GETARG_UINT64(1);
	uint64 result;

	result = (arg1 > arg2) ? arg1 : arg2;

	PG_RETURN_UINT64(result);
}

PG_FUNCTION_INFO_V1(int2int4mod);
Datum
int2int4mod(PG_FUNCTION_ARGS)
{
	int16 arg1 = PG_GETARG_INT16(0);
	int32 arg2 = PG_GETARG_INT32(1);
	int32 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	if (arg2 == -1)
		PG_RETURN_INT32(0);

	result = arg1 % arg2;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int4int2mod);
Datum
int4int2mod(PG_FUNCTION_ARGS)
{
	int32 arg1 = PG_GETARG_INT32(0);
	int16 arg2 = PG_GETARG_INT16(1);
	int32 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	if (arg2 == -1)
		PG_RETURN_INT32(0);

	result = arg1 % arg2;

	PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(int8int2mod);
Datum
int8int2mod(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	int16 arg2 = PG_GETARG_INT16(1);
	int64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	if (arg2 == -1)
		PG_RETURN_INT64(0);

	result = arg1 % arg2;

	PG_RETURN_INT64(result);
}

PG_FUNCTION_INFO_V1(int8int4mod);
Datum
int8int4mod(PG_FUNCTION_ARGS)
{
	int64 arg1 = PG_GETARG_INT64(0);
	int32 arg2 = PG_GETARG_INT32(1);
	int64 result;

	if (arg2 == 0)
	{
		ereport(ERROR,
			(errcode(ERRCODE_DIVISION_BY_ZERO),
			 errmsg("division by zero")));
		PG_RETURN_NULL();
	}

	if (arg2 == -1)
		PG_RETURN_INT64(0);

	result = arg1 % arg2;

	PG_RETURN_INT64(result);
}
